/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###CLASS###
Name:      Boing
Version:   1.0
Status:    Stable
Date:      July 2001
ID:        ID_BOING
Category:  Effect
Author:    Rocklyte Systems
Copyright: Rocklyte Systems (c) 2000-2001.  All rights reserved.
Short:     Real-time animated bouncing circles.
Keywords:  boing, bounce, circle, ellipse, ball

###DESCRIPTION###
<p>The Boing class draws a number of multi-coloured circles that bounce
around the sides and bottom of the graphical container.  The class draws to
the foreground only, so you will need to use a graphical object to define
the background underneath the Boing effect.  Here is an example:</p>

<pre>
  &lt;render width="300" height="200"/&gt;
    &lt;box colour="#000000"/&gt;
    &lt;boing/&gt;
  &lt;/render&gt;
</pre>

<p>Once initialised, the effect will run automatically until the Boing
object is freed from the system.  The amount of processor cycles used by
the Boing object is directly proportional to the amount of bouncing
elements, as specified by the MaxElements field.  Generally, 50 to 100
elements are sufficient for an adequate effect and minimal processor
usage.</p>

###END###

*****************************************************************************/

#include <pandora/system/all.h>
#include <pandora/graphics/bitmap.h>
#include <pandora/graphics/bob.h>
#include <pandora/graphics/render.h>
#include <math.h>

static ERROR CMDInit(OBJECTPTR, struct KernelBase *);
static ERROR CMDExpunge(void);

#define VER_BOING 1.0

EXPORT struct ModHeader ModHeader = {
   MODULE_HEADER_V1, CMDInit, NULL, NULL, CMDExpunge,
   JMP_DEFAULT, 0, CPU_PC, VER_BOING, VER_KERNEL,
   "Rocklyte Systems",
   "Rocklyte Systems (c) 2000-2001.  All rights reserved.",
   "July 2001",
   "Boing"
};

static struct KernelBase *KernelBase = 0;
static OBJECTPTR BoingClass = 0;

/*****************************************************************************
** Boing class.
*/

struct Boing {
   OBJECT_HEADER
   OBJECTID DrawableID; /* Reference to the drawable to use for the Boing graphics */
   LONG MaxElements;    /* Maximum amount of stars to use */
   FLOAT Gravity;        /* The amount of gravity to be applied */
   OBJECTID BobID;      /* Reference to image graphics, if an image should be used */

   /*** Field store ***/
   OBJECTID PenID;
   MEMORYID BallsMID;
   struct Ball *Balls;
   struct BitmapPen *Pen;
};

struct Ball {
   FLOAT X;
   FLOAT Y;
   WORD Frame;
   FLOAT XSpeed;
   FLOAT YSpeed;
   WORD Width;
   WORD Height;
   WORD Fill;
   struct RGB Colour;
};

static struct FieldArray BoingFields[] = {
   { "Drawable",    0, FDF_OBJECTID|FDF_R,  0, NULL, NULL },
   { "MaxElements", 0, FDF_LONG|FDF_RI,     0, NULL, NULL },
   { "Gravity",     0, FDF_FLOAT|FDF_RI,     0, NULL, NULL },
   { "Bob",         0, FDF_OBJECTID|FDF_RI, 0, NULL, NULL },
   END_FIELD
};

static ERROR BOING_ActionNotify(struct Boing *, struct acActionNotify *);
static ERROR BOING_Free(struct Boing *, APTR);
static ERROR BOING_ReleaseObject(struct Boing *, APTR);
static ERROR BOING_Init(struct Boing *, APTR);
static ERROR BOING_NewObject(struct Boing *, APTR);
static ERROR BOING_Timer(struct Boing *, struct acTimer *);

static struct ActionArray BoingActions[] = {
   { AC_ActionNotify,  BOING_ActionNotify },
   { AC_Free,          BOING_Free },
   { AC_ReleaseObject, BOING_ReleaseObject },
   { AC_Init,          BOING_Init },
   { AC_NewObject,     BOING_NewObject },
   { AC_Timer,         BOING_Timer },
   { NULL, NULL }
};

/*****************************************************************************
** Command: Init()
*/

static ERROR CMDInit(OBJECTPTR argModule, struct KernelBase *argKernelBase)
{
   KernelBase = argKernelBase;

   return(CreateObject(ID_CLASS, NULL, (OBJECTPTR *)&BoingClass, NULL,
      FID_BaseClassID|TLONG, ID_BOING,
      FID_Version|TFLOAT,    VER_BOING,
      FID_Name|TSTRING,      "Boing",
      FID_Category|TLONG,    CCF_EFFECT,
      FID_Actions|TPTR,      BoingActions,
      FID_Fields|TPTR,       BoingFields,
      FID_Size|TLONG,        sizeof(struct Boing),
      TAGEND));
}

/*****************************************************************************
** Command: Expunge()
*/

static ERROR CMDExpunge(void)
{
   if (BoingClass) { Action(AC_Free, BoingClass, NULL); BoingClass = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Boing: ActionNotify
*/

static ERROR BOING_ActionNotify(struct Boing *Self, struct acActionNotify *Args)
{
   if ((Args->ActionID IS AC_Draw) AND (Args->Error IS ERR_Okay)) {
      struct Bitmap *Bitmap;
      struct Bob *bob;
      OBJECTPTR Drawable;
      LONG i;

      if (AccessObject(Self->DrawableID, 1000, &Drawable) != ERR_Okay) {
         return(ObjectError(Self, ERH_Draw, ERR_ExclusiveDenied));
      }

      if ((GetField(Drawable, FID_Bitmap, FT_POINTER, &Bitmap) != ERR_Okay) OR (!Bitmap)) {
         ReleaseObject(Drawable);
         return(ObjectError(Self, ERH_Draw, ERR_Failed));
      }

      if (!Self->Balls) AccessMemory(Self->BallsMID, MEM_READWRITE, (void **)&Self->Balls);
      if (!Self->Pen)   AccessObject(Self->PenID, 0, (OBJECTPTR *)&Self->Pen);

      if ((Self->Balls) AND (Self->Pen)) {
         if (Self->BobID) {
            if (AccessObject(Self->BobID, 3000, (OBJECTPTR *)&bob) IS ERR_Okay) {
               for (i=0; i < Self->MaxElements; i++) {
                  bob->XCoord = Self->Balls[i].X;
                  bob->YCoord = Self->Balls[i].Y;
                  bob->Frame  = Self->Balls[i].Frame;
                  bob->Destination = Bitmap;
                  Action(AC_Draw, bob, NULL);
               }
               ReleaseObject(bob);
            }
         }
         else {
            /* Blit circles if we don't have a bob ***/
            for (i=0; i < Self->MaxElements; i++) {
               Self->Pen->Bitmap   = Bitmap;
               Self->Pen->Red      = Self->Balls[i].Colour.Red;
               Self->Pen->Green    = Self->Balls[i].Colour.Green;
               Self->Pen->Blue     = Self->Balls[i].Colour.Blue;
               Self->Pen->XCoord   = Self->Balls[i].X;
               Self->Pen->YCoord   = Self->Balls[i].Y;
               Self->Pen->Width    = Self->Balls[i].Width;
               Self->Pen->Height   = Self->Balls[i].Height;
               Self->Pen->Fill     = Self->Balls[i].Fill;
               Self->Pen->TipShape = TIP_PIXEL;
               Self->Pen->PenShape = PEN_ELLIPSE;
               Self->Pen->TipSize  = 1;
               Action(AC_Draw, Self->Pen, NULL);
            }
         }
      }

      ReleaseObject(Drawable);
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Balls: Free
*/

static ERROR BOING_Free(struct Boing *Self, APTR Void)
{
   UnsubscribeTimer(Self->Head.UniqueID);

   if (Self->DrawableID) {
      OBJECTPTR Drawable;
      if (AccessObject(Self->DrawableID, 2000, &Drawable) IS ERR_Okay) {
         UnsubscribeAction(Drawable, NULL, GetUniqueID(Self));
         ReleaseObject(Drawable);
      }
      Self->DrawableID = NULL;
   }

   if (Self->Balls)    { ReleaseMemory(Self->Balls); Self->Balls = NULL; }
   if (Self->BallsMID) { FreeMemoryID(Self->BallsMID); Self->BallsMID = NULL; }

   if (Self->Pen)   { ReleaseObject(Self->Pen); Self->Pen = NULL; }
   if (Self->PenID) { ActionMsg(AC_Free, Self->PenID, NULL); Self->PenID = NULL; }

   return(ERR_Okay);
}

/*****************************************************************************
** Boing: ReleaseObject()
*/

static ERROR BOING_ReleaseObject(struct Boing *Self, APTR Void)
{
   if (Self->Balls) { ReleaseMemory(Self->Balls); Self->Balls = NULL; }
   if (Self->Pen)   { ReleaseObject(Self->Pen); Self->Pen = NULL; }
   return(ERR_Okay);
}

/*****************************************************************************
** Boing: Init
*/

static ERROR BOING_Init(struct Boing *Self, APTR Void)
{
   OBJECTID OwnerID;
   OBJECTPTR Drawable;
   struct Bob *bob;
   LONG DrawableWidth, DrawableHeight, i;

   if (Self->MaxElements < 1) Self->MaxElements = 50;
   if (Self->MaxElements > 500) Self->MaxElements = 500;

   /*** Find the Drawable object that we are associated with ***/

   if (!Self->DrawableID) {
      OwnerID = GetOwner(Self);
      while ((OwnerID) AND (GetClassID(OwnerID) != ID_RENDER)) {
         OwnerID = GetOwnerID(OwnerID);
      }
      if (!OwnerID) return(ObjectError(Self, ERH_Init, ERR_UnsupportedOwner));
      else Self->DrawableID = OwnerID;
   }

   if (AccessObject(Self->DrawableID, 2000, &Drawable) IS ERR_Okay) {
      if (GetFields(Drawable, FID_Width|TLONG,  &DrawableWidth,
                              FID_Height|TLONG, &DrawableHeight,
                              TAGEND) IS ERR_Okay) {
         if (AllocMemory(sizeof(struct Ball) * Self->MaxElements, Self->Head.MemFlags|MEM_NOCLEAR, (APTR *)&Self->Balls, &Self->BallsMID) IS ERR_Okay) {
            if (Self->BobID) {
               if (AccessObject(Self->BobID, 3000, (OBJECTPTR *)&bob) IS ERR_Okay) {
                  for (i=0; i < Self->MaxElements; i++) {
                     Self->Balls[i].Frame  = RandomNumber(bob->AmtFrames);
                     Self->Balls[i].Width  = bob->Frames[Self->Balls[i].Frame].Width;
                     Self->Balls[i].Height = bob->Frames[Self->Balls[i].Frame].Height;
                     Self->Balls[i].X      = RandomNumber(DrawableWidth - Self->Balls[i].Width);
                     Self->Balls[i].Y      = RandomNumber(DrawableHeight - Self->Balls[i].Height);
                     Self->Balls[i].XSpeed = ((FLOAT)RandomNumber(700000)+400000)/100000;
                     Self->Balls[i].YSpeed = ((FLOAT)RandomNumber(800000)+300000)/100000;
                     if (Self->Balls[i].Frame & 1) Self->Balls[i].XSpeed = -Self->Balls[i].XSpeed;
                  }
                  ReleaseObject(bob);
               }
            }
            else {
               for (i=0; i < Self->MaxElements; i++) {
                  Self->Balls[i].Frame  = 0;
                  Self->Balls[i].Width  = 5 + RandomNumber(40);
                  Self->Balls[i].Height = Self->Balls[i].Width;
                  Self->Balls[i].X      = RandomNumber(DrawableWidth - Self->Balls[i].Width);
                  Self->Balls[i].Y      = RandomNumber(DrawableHeight - Self->Balls[i].Height);
                  Self->Balls[i].XSpeed = ((FLOAT)RandomNumber(700000)+400000)/100000;
                  Self->Balls[i].YSpeed = ((FLOAT)RandomNumber(800000)+300000)/100000;
                  Self->Balls[i].Fill   = RandomNumber(2);
                  Self->Balls[i].Colour.Red   = 50 + RandomNumber(200);
                  Self->Balls[i].Colour.Green = 50 + RandomNumber(200);
                  Self->Balls[i].Colour.Blue  = 50 + RandomNumber(200);
               }
            }
         }
      }
      ReleaseObject(Drawable);
   }
   else return(ObjectError(Self, ERH_Init, ERR_ExclusiveDenied));

   /*** Action subscriptions ***/

   if (AccessObject(Self->DrawableID, 200, &Drawable) IS ERR_Okay) {
      SubscribeAction(Drawable, AC_Draw, Self);
      ReleaseObject(Drawable);
   }
   else return(ObjectError(Self, ERH_Init, ERR_ExclusiveDenied));

   /*** Subscribe to the Timer ***/

   return(SubscribeTimer(Self->Head.UniqueID, 30));
}

/*****************************************************************************
** Boing: NewObject
*/

static ERROR BOING_NewObject(struct Boing *Self, APTR Void)
{
   if (NewObject(ID_PEN, Self->Head.ObjectFlags|NF_CHILD, &Self->Pen, &Self->PenID) IS ERR_Okay) {
      if (Action(AC_Init, Self->Pen, NULL) IS ERR_Okay) {
         Self->MaxElements = 100;
         Self->Gravity     = 1.0;
         return(ERR_Okay);
      }
      else return(ObjectError(Self, ERH_NewObject, ERR_Init));
   }
   else return(ObjectError(Self, ERH_NewObject, ERR_NewObject));
}

/*****************************************************************************
** Boing: Timer
*/

static ERROR BOING_Timer(struct Boing *Self, struct acTimer *Args)
{
   OBJECTPTR Drawable;
   LONG DrawableWidth, DrawableHeight, i;

   if (AccessObject(Self->DrawableID, 2000, &Drawable) IS ERR_Okay) {
      if (!Self->Balls) {
         if (AccessMemory(Self->BallsMID, MEM_READWRITE, (void **)&Self->Balls) != ERR_Okay) {
            ReleaseObject(Drawable);
            return(ERR_ExclusiveDenied);
         }
      }

      if (GetFields(Drawable, FID_Width|TLONG,  &DrawableWidth,
                              FID_Height|TLONG, &DrawableHeight,
                              TAGEND) IS ERR_Okay) {

         for (i=0; i < Self->MaxElements; i++) {
            if ((Self->Balls[i].Y + Self->Balls[i].YSpeed + Self->Balls[i].Height) > DrawableHeight) {
               if (Self->Balls[i].YSpeed > 0)   Self->Balls[i].YSpeed = -Self->Balls[i].YSpeed;
               if (Self->Balls[i].YSpeed < -30) Self->Balls[i].YSpeed = -(RandomNumber(20)+10);
               if (Self->Balls[i].YSpeed > -5)  Self->Balls[i].YSpeed = -(RandomNumber(20)+10);
               Self->Balls[i].YSpeed += Self->Gravity;
               Self->Balls[i].Y = DrawableHeight - Self->Balls[i].Height;
            }
            else {
               Self->Balls[i].Y += Self->Balls[i].YSpeed;
               Self->Balls[i].YSpeed += Self->Gravity;
            }

            if (Self->Balls[i].Y + Self->Balls[i].Height > DrawableHeight) Self->Balls[i].Y = RandomNumber(DrawableHeight - Self->Balls[i].Height);

            if ((Self->Balls[i].X + Self->Balls[i].XSpeed) < 0) {
               if (Self->Balls[i].XSpeed < 0) Self->Balls[i].XSpeed = -Self->Balls[i].XSpeed;
               Self->Balls[i].X += Self->Balls[i].XSpeed;
            }
            else if ((Self->Balls[i].X + Self->Balls[i].XSpeed + Self->Balls[i].Width) >= DrawableWidth) {
               if (Self->Balls[i].XSpeed > 0) Self->Balls[i].XSpeed = -Self->Balls[i].XSpeed;
               Self->Balls[i].X += Self->Balls[i].XSpeed;
            }
            else Self->Balls[i].X += Self->Balls[i].XSpeed;
         }
      }
      ReleaseObject(Drawable);
   }

   DelayMsg(AC_Draw, Self->DrawableID, NULL);

   return(ERR_Okay);
}

/*****************************************************************************

###FIELD###
Name:   MaxElements
Short:  The total number of bouncing elements is defined by this field.
Type:   LONG
Status: Read/Init

###DESCRIPTION###
<p>You may specify the exact amount of elements to be used in the Boing effect
by setting the MaxElements field.  For a reasonable effect, a minimum amount
of 50 elements is recommended.  The value should be no higher than 300.</p>

<p>If you do not set this field, the default number of elements will be used.</p>

###END###

*****************************************************************************/
